import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2


class FeatureAssociation(APIObject):
    """ Feature association statistics for a project.

    Attributes
    ----------
    type : str
        Either 'association' or 'correlation' the class of the pairwise stats
    metric : str
        the metric of either class of pairwise stats
        'spearman', 'pearson', etc for correlation, 'mutualInfo', 'cramersV'
        for association
    """

    _converter = t.Dict(
        {
            t.Key("type"): t.Or(t.String, t.Null),
            t.Key("metric"): t.Or(t.String, t.Null),
            t.Key("featurelistId"): t.Or(t.String, t.Null),
        }
    )

    def __init__(self, metric=None, assoc_type=None, featurelistId=None):
        self.metric = metric if metric else "mutualInfo"
        self.assoc_type = type if type else "association"
        self.featurelistId = featurelistId

    def __repr__(self):
        return encode_utf8_if_py2(
            u"FeatureAssociation({}/{}/{})".format(self.assoc_type, self.metric, self.featurelistId)
        )


class FeatureAssociationMatrixDetails(APIObject):
    """ Plotting details for a pair of passed features present in the feature
    association matrix

    Attributes
    ----------
    feature1 : str
        Feature name for the first feature of interest
    feature2 : str
        Feature name for the second feature of interest
    """

    _converter = t.Dict({t.Key("feature1"): t.String, t.Key("feature2"): t.String})

    def __init__(self, feature1=None, feature2=None):
        self.feature1 = feature1
        self.feature2 = feature2

    def __repr__(self):
        return encode_utf8_if_py2(
            u"FeatureAssociationMatrixDetails({}/{})".format(self.feature1, self.feature2)
        )


class FeatureAssociationFeaturelists(APIObject):
    """ Get project featurelists and see if they have association statistics
    """

    def __repr__(self):
        return encode_utf8_if_py2(u"FeatureAssocationFeaturelists")



import numpy as np
import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2


class RocCurveThresholdMixin(object):
    roc_points = None

    @staticmethod
    def _validate_threshold(threshold):
        if threshold > 1 or threshold < 0:
            raise ValueError("threshold must be from [0, 1] interval")

    def estimate_threshold(self, threshold):
        """ Return metrics estimation for given threshold.

        Parameters
        ----------
        threshold : float from [0, 1] interval
            Threshold we want estimation for

        Returns
        -------
        dict
            Dictionary of estimated metrics in form of {metric_name: metric_value}.
            Metrics are 'accuracy', 'f1_score', 'false_negative_score', 'true_negative_score',
            'true_negative_rate', 'matthews_correlation_coefficient', 'true_positive_score',
            'positive_predictive_value', 'false_positive_score', 'false_positive_rate',
            'negative_predictive_value', 'true_positive_rate'.

        Raises
        ------
        ValueError
            Given threshold isn't from [0, 1] interval
        """
        self._validate_threshold(threshold)
        for roc_point in self.roc_points:
            if np.isclose(roc_point["threshold"], threshold):
                estimated_roc_point = roc_point
                break
        else:
            # if no exact match - pick closest ROC point with bigger threshold
            roc_points_with_bigger_threshold = [
                roc_point for roc_point in self.roc_points if roc_point["threshold"] > threshold
            ]
            estimated_roc_point = sorted(
                roc_points_with_bigger_threshold, key=lambda rp: rp["threshold"]
            )[0]
        return estimated_roc_point

    def get_best_f1_threshold(self):
        """ Return value of threshold that corresponds to max F1 score.
        This is threshold that will be preselected in DataRobot when you open "ROC curve" tab.

        Returns
        -------
        float
            Threhold with best F1 score.
        """
        roc_point_with_best_f1 = max(self.roc_points, key=lambda roc_point: roc_point["f1_score"])
        return roc_point_with_best_f1["threshold"]


RocPointsTrafaret = t.Dict(
    {
        t.Key("negative_class_predictions"): t.List(t.Float),
        t.Key("positive_class_predictions"): t.List(t.Float),
        t.Key("roc_points"): t.List(
            t.Dict(
                {
                    t.Key("accuracy"): t.Float,
                    t.Key("f1_score"): t.Float,
                    t.Key("false_negative_score"): t.Int,
                    t.Key("true_negative_score"): t.Int,
                    t.Key("true_positive_score"): t.Int,
                    t.Key("false_positive_score"): t.Int,
                    t.Key("true_negative_rate"): t.Float,
                    t.Key("false_positive_rate"): t.Float,
                    t.Key("true_positive_rate"): t.Float,
                    t.Key("matthews_correlation_coefficient"): t.Float,
                    t.Key("positive_predictive_value"): t.Float,
                    t.Key("negative_predictive_value"): t.Float,
                    t.Key("threshold"): t.Float,
                    t.Key("fraction_predicted_as_positive"): t.Float,
                    t.Key("fraction_predicted_as_negative"): t.Float,
                    t.Key("lift_positive"): t.Float,
                    t.Key("lift_negative"): t.Float,
                }
            ).ignore_extra("*")
        ),
    }
)


class RocCurve(APIObject, RocCurveThresholdMixin):
    """ ROC curve data for model.

    Attributes
    ----------
    source : str
        ROC curve data source. Can be 'validation', 'crossValidation' or 'holdout'.
    roc_points : list of dict
        List of precalculated metrics associated with thresholds for ROC curve.
    negative_class_predictions : list of float
        List of predictions from example for negative class
    positive_class_predictions : list of float
        List of predictions from example for positive class
    source_model_id : str
        ID of the model this ROC curve represents; in some cases,
        insights from the parent of a frozen model may be used
    """

    _converter = (
        t.Dict({t.Key("source"): t.String, t.Key("source_model_id"): t.String})
        .merge(RocPointsTrafaret)
        .ignore_extra("*")
    )

    def __init__(
        self,
        source,
        roc_points,
        negative_class_predictions,
        positive_class_predictions,
        source_model_id,
    ):
        self.source = source
        self.roc_points = roc_points
        self.negative_class_predictions = negative_class_predictions
        self.positive_class_predictions = positive_class_predictions
        self.source_model_id = source_model_id

    def __repr__(self):
        return encode_utf8_if_py2(u"RocCurve({})".format(self.source))



# flake8: noqa

from ._version import __version__
from .client import Client
from .enums import AUTOPILOT_MODE, QUEUE_STATUS, SCORING_TYPE, TARGET_TYPE, VERBOSITY_LEVEL
from .errors import AppPlatformError
from .helpers import *
from .models import (
    BatchPredictionJob,
    BlenderModel,
    Blueprint,
    BlueprintChart,
    BlueprintTaskDocument,
    CalendarFile,
    ComplianceDocTemplate,
    ComplianceDocumentation,
    Credential,
    CustomInferenceImage,
    CustomInferenceModel,
    CustomModelTest,
    CustomModelVersion,
    CustomModelVersionDependencyBuild,
    DataDriver,
    Dataset,
    DatasetDetails,
    DatasetFeature,
    DatasetFeatureHistogram,
    DatasetFeaturelist,
    DataSource,
    DataSourceParameters,
    DataStore,
    DatetimeModel,
    Deployment,
    ExecutionEnvironment,
    ExecutionEnvironmentVersion,
    ExternalConfusionChart,
    ExternalLiftChart,
    ExternalMulticlassLiftChart,
    ExternalResidualsChart,
    ExternalRocCurve,
    ExternalScores,
    Feature,
    FeatureEngineeringGraph,
    FeatureHistogram,
    FeatureImpactJob,
    Featurelist,
    FrozenModel,
    ImportedModel,
    InteractionFeature,
    Job,
    Model,
    ModelBlueprintChart,
    ModelingFeature,
    ModelingFeaturelist,
    ModelJob,
    ModelRecommendation,
    PayoffMatrix,
    PredictionDataset,
    PredictionExplanations,
    PredictionExplanationsInitialization,
    Predictions,
    PredictionServer,
    PredictJob,
    PrimeFile,
    PrimeModel,
    Project,
    RatingTable,
    RatingTableModel,
    ReasonCodes,
    ReasonCodesInitialization,
    RelationshipsConfiguration,
    Ruleset,
    SecondaryDatasetConfigurations,
    ShapImpact,
    ShapMatrix,
    ShapMatrixJob,
    SharingAccess,
    TrainingPredictions,
    TrainingPredictionsJob,
)



import json

import pytest
import responses

from datarobot import Project


@pytest.fixture
def project(project_with_target_data):
    return Project.from_data(project_with_target_data)


@pytest.fixture
def feature_association_details_data():
    return {
        "values": [
            [168, 33, 0.045],
            [1, 1, 0.329],
            [37, 17, 0.005],
            [207, 41, 0.014],
            [178, 35, 0.006],
            [34, 16, 0.032],
            [194, 37, 0.004],
        ],
        "features": ["area1", "area2"],
        "types": ["N", "N"],
    }


@pytest.fixture
def feature_association_featurelists():
    return {
        "featurelists": [
            {
                "featurelistId": "54e510ef8bd88f5aeb02a3ed",
                "hasFam": True,
                "title": "Informative Features",
            }
        ]
    }


@pytest.fixture
def feature_association_list_url(project_id):
    base_url = "https://host_name.com/projects/{}/featureAssociationMatrixDetails/".format(
        project_id
    )
    base_url += "?feature1=area1&feature2=area2"
    return base_url


@pytest.fixture
def feature_association_featurelist_url(project_id):
    url = "https://host_name.com/projects/{}/featureAssociationFeaturelists/".format(project_id)
    return url


@responses.activate
def test_get_feature_association_details(
    feature_association_details_data, feature_association_list_url, project, project_id
):
    responses.add(
        responses.GET,
        feature_association_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps(feature_association_details_data),
    )
    feature_values = project.get_association_matrix_details(feature1="area1", feature2="area2")

    assert len(feature_values.keys()) == 3
    assert len(feature_values["values"]) == 7
    assert feature_values["features"] == ["area1", "area2"]


@responses.activate
def test_get_feature_association_featurelists(
    feature_association_featurelists, feature_association_featurelist_url, project, project_id
):
    responses.add(
        responses.GET,
        feature_association_featurelist_url,
        status=200,
        content_type="application/json",
        body=json.dumps(feature_association_featurelists),
    )
    featurelists = project.get_association_featurelists()

    assert len(featurelists.keys()) == 1
    assert len(featurelists["featurelists"]) == 1
    assert featurelists["featurelists"][0]["title"] == "Informative Features"



import json

import pytest
import responses

from datarobot import Model
from datarobot.errors import ClientError
from datarobot.models.roc_curve import RocCurve
from datarobot.utils import from_api


@pytest.fixture
def roc_curve_validation_data(model_id, roc_curve_data):
    return {
        "source": "validation",
        "sourceModelId": model_id,
        "rocPoints": roc_curve_data["rocPoints"],
        "negativeClassPredictions": roc_curve_data["negativeClassPredictions"],
        "positiveClassPredictions": roc_curve_data["positiveClassPredictions"],
    }


@pytest.fixture
def roc_curve_parent_model_validation_data(parent_model_id, roc_curve_data):
    return {
        "source": "validation",
        "sourceModelId": parent_model_id,
        "rocPoints": roc_curve_data["rocPoints"],
        "negativeClassPredictions": roc_curve_data["negativeClassPredictions"],
        "positiveClassPredictions": roc_curve_data["positiveClassPredictions"],
    }


@pytest.fixture
def roc_curve_parent_model_holdout_data(parent_model_id, roc_curve_data):
    return {
        "source": "holdout",
        "sourceModelId": parent_model_id,
        "rocPoints": roc_curve_data["rocPoints"],
        "negativeClassPredictions": roc_curve_data["negativeClassPredictions"],
        "positiveClassPredictions": roc_curve_data["positiveClassPredictions"],
    }


def test_instantiation(roc_curve_validation_data):
    roc = RocCurve.from_server_data(roc_curve_validation_data)

    assert roc.source == roc_curve_validation_data["source"]
    assert roc.negative_class_predictions == roc_curve_validation_data["negativeClassPredictions"]
    assert roc.positive_class_predictions == roc_curve_validation_data["positiveClassPredictions"]
    assert roc.roc_points == from_api(roc_curve_validation_data["rocPoints"])
    assert roc.source_model_id == roc_curve_validation_data["sourceModelId"]


def test_future_proof(roc_curve_validation_data):
    data_with_future_keys = dict(roc_curve_validation_data, new_key="some future roc data")
    data_with_future_keys["rocPoints"][0]["new_key"] = "some future bin data"
    RocCurve.from_server_data(data_with_future_keys)


@pytest.fixture
def roc_curve_validation_data_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/rocCurve/validation/".format(
        project_id, model_id
    )


@pytest.fixture
def roc_curve_parent_model_validation_data_url(project_id, parent_model_id):
    return "https://host_name.com/projects/{}/models/{}/rocCurve/validation/".format(
        project_id, parent_model_id
    )


@pytest.fixture
def roc_curve_list_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/rocCurve/".format(project_id, model_id)


@pytest.fixture
def roc_curve_parent_list_url(project_id, parent_model_id):
    return "https://host_name.com/projects/{}/models/{}/rocCurve/".format(
        project_id, parent_model_id
    )


@pytest.fixture
def frozen_model_url(project_id, model_id):
    return "https://host_name.com/projects/{}/frozenModels/{}/".format(project_id, model_id)


@responses.activate
def test_get_validation_roc_curve(
    roc_curve_validation_data, roc_curve_validation_data_url, project_id, model_id
):
    responses.add(
        responses.GET,
        roc_curve_validation_data_url,
        status=200,
        content_type="application/json",
        body=json.dumps(roc_curve_validation_data),
    )
    model = Model(id=model_id, project_id=project_id)
    roc = model.get_roc_curve("validation")

    assert roc.source == roc_curve_validation_data["source"]
    assert roc.negative_class_predictions == roc_curve_validation_data["negativeClassPredictions"]
    assert roc.positive_class_predictions == roc_curve_validation_data["positiveClassPredictions"]
    assert roc.roc_points == from_api(roc_curve_validation_data["rocPoints"])
    assert roc.source_model_id == model_id


@responses.activate
def test_get_frozen_validation_roc_curve_no_fallback(
    roc_curve_validation_data_url, project_id, model_id
):
    responses.add(responses.GET, roc_curve_validation_data_url, status=404)
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    with pytest.raises(ClientError):
        model.get_roc_curve("validation")


@responses.activate
@pytest.mark.usefixtures("known_warning")
def test_get_frozen_validation_roc_curve_with_fallback(
    roc_curve_parent_model_validation_data,
    roc_curve_validation_data_url,
    roc_curve_parent_model_validation_data_url,
    frozen_model_url,
    project_id,
    model_id,
    parent_model_id,
    frozen_json,
):
    responses.add(responses.GET, roc_curve_validation_data_url, status=404)
    responses.add(
        responses.GET,
        frozen_model_url,
        body=frozen_json,
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        roc_curve_parent_model_validation_data_url,
        status=200,
        content_type="application/json",
        body=json.dumps(roc_curve_parent_model_validation_data),
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    roc = model.get_roc_curve("validation", fallback_to_parent_insights=True)

    assert roc.source == roc_curve_parent_model_validation_data["source"]
    assert (
        roc.negative_class_predictions
        == roc_curve_parent_model_validation_data["negativeClassPredictions"]
    )
    assert (
        roc.positive_class_predictions
        == roc_curve_parent_model_validation_data["positiveClassPredictions"]
    )
    assert roc.roc_points == from_api(roc_curve_parent_model_validation_data["rocPoints"])
    assert roc.source_model_id == parent_model_id


@responses.activate
def test_get_all_roc_curves(roc_curve_validation_data, roc_curve_list_url, project_id, model_id):
    responses.add(
        responses.GET,
        roc_curve_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps({"charts": [roc_curve_validation_data]}),
    )
    model = Model(id=model_id, project_id=project_id)
    roc_list = model.get_all_roc_curves()

    assert len(roc_list) == 1
    assert roc_list[0].source == roc_curve_validation_data["source"]
    assert (
        roc_list[0].negative_class_predictions
        == roc_curve_validation_data["negativeClassPredictions"]
    )
    assert (
        roc_list[0].positive_class_predictions
        == roc_curve_validation_data["positiveClassPredictions"]
    )
    assert roc_list[0].roc_points == from_api(roc_curve_validation_data["rocPoints"])
    assert roc_list[0].source_model_id == model_id


@responses.activate
def test_get_frozen_all_roc_curves_no_fallback(
    roc_curve_validation_data, roc_curve_list_url, project_id, model_id
):
    responses.add(
        responses.GET,
        roc_curve_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps({"charts": [roc_curve_validation_data]}),
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    roc_list = model.get_all_roc_curves()

    assert len(roc_list) == 1
    assert roc_list[0].source == roc_curve_validation_data["source"]
    assert (
        roc_list[0].negative_class_predictions
        == roc_curve_validation_data["negativeClassPredictions"]
    )
    assert (
        roc_list[0].positive_class_predictions
        == roc_curve_validation_data["positiveClassPredictions"]
    )
    assert roc_list[0].roc_points == from_api(roc_curve_validation_data["rocPoints"])
    assert roc_list[0].source_model_id == model_id


@responses.activate
@pytest.mark.usefixtures("known_warning")
def test_get_forzen_all_roc_curves_with_fallback(
    roc_curve_validation_data,
    roc_curve_parent_model_validation_data,
    roc_curve_parent_model_holdout_data,
    roc_curve_list_url,
    roc_curve_parent_list_url,
    frozen_model_url,
    project_id,
    model_id,
    parent_model_id,
    frozen_json,
):
    responses.add(
        responses.GET,
        roc_curve_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps({"charts": [roc_curve_validation_data]}),
    )
    responses.add(
        responses.GET,
        frozen_model_url,
        body=frozen_json,
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        roc_curve_parent_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps(
            {
                "charts": [
                    roc_curve_parent_model_validation_data,
                    roc_curve_parent_model_holdout_data,
                ]
            }
        ),
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    roc_list = model.get_all_roc_curves(fallback_to_parent_insights=True)

    assert len(roc_list) == 2
    assert roc_list[0].source == roc_curve_validation_data["source"]
    assert (
        roc_list[0].negative_class_predictions
        == roc_curve_validation_data["negativeClassPredictions"]
    )
    assert (
        roc_list[0].positive_class_predictions
        == roc_curve_validation_data["positiveClassPredictions"]
    )
    assert roc_list[0].roc_points == from_api(roc_curve_validation_data["rocPoints"])
    assert roc_list[0].source_model_id == model_id

    assert roc_list[1].source == roc_curve_parent_model_holdout_data["source"]
    assert (
        roc_list[1].negative_class_predictions
        == roc_curve_parent_model_holdout_data["negativeClassPredictions"]
    )
    assert (
        roc_list[1].positive_class_predictions
        == roc_curve_parent_model_holdout_data["positiveClassPredictions"]
    )
    assert roc_list[1].roc_points == from_api(roc_curve_parent_model_holdout_data["rocPoints"])
    assert roc_list[1].source_model_id == parent_model_id


def test_get_best_f1_threshold(roc_curve_validation_data):
    # Given ROC curve data
    roc = RocCurve.from_server_data(roc_curve_validation_data)
    # When calculate recommended threshold
    best_threshold = roc.get_best_f1_threshold()
    # Then f1 score for that threshold maximal from all ROC curve points
    best_f1 = roc.estimate_threshold(best_threshold)["f1_score"]
    assert all(best_f1 >= roc_point["f1_score"] for roc_point in roc.roc_points)


def test_estimate_threshold_equal(roc_curve_validation_data):
    # Given ROC curve data
    roc = RocCurve.from_server_data(roc_curve_validation_data)
    # When estimating threshold equal to one of precalculated
    threshold = roc_curve_validation_data["rocPoints"][1]["threshold"]
    # Then estimate_threshold return valid data
    assert roc.estimate_threshold(threshold)["threshold"] == threshold


def test_estimate_threshold_new(roc_curve_validation_data):
    # Given ROC curve data
    roc = RocCurve.from_server_data(roc_curve_validation_data)
    # When estimating threshold from outside of precalculated
    threshold = roc_curve_validation_data["rocPoints"][1]["threshold"] + 0.1
    # Then estimate_threshold return data for next threshold bigger then requested
    assert roc.estimate_threshold(threshold)["threshold"] > threshold



import json

import pytest
import responses

from datarobot import errors
from datarobot.models.secondary_dataset import (
    DatasetConfiguration,
    SecondaryDataset,
    SecondaryDatasetConfigurations,
)
from tests.utils import request_body_to_json


@pytest.fixture
def ids():
    return {
        "pid": "5a530498d5c1f302d6d17699",
        "graph_id": "5a530498d5c1f302d6d17111",
        "identifier_1": "test_dataset_1",
        "catalog_version_id_1": "5a530498d5c1f302d6d17121",
        "catalog_id_1": "5a530498d5c1f302d6d17115",
        "identifier_2": "test_dataset_2",
        "catalog_version_id_2": "6a530498d5c1f302d6d17121",
        "catalog_id_2": "6a530498d5c1f302d6d17115",
    }


@pytest.fixture
def dataset_configurations(ids):
    secondary_dataset_1 = SecondaryDataset(
        identifier=ids["identifier_1"],
        catalog_version_id=ids["catalog_version_id_1"],
        catalog_id=ids["catalog_id_1"],
        snapshot_policy="specified",
    )
    secondary_dataset_2 = SecondaryDataset(
        identifier=ids["identifier_2"],
        catalog_version_id=ids["catalog_version_id_2"],
        catalog_id=ids["catalog_id_2"],
        snapshot_policy="specified",
    )
    return [
        DatasetConfiguration(
            feature_engineering_graph_id=ids["graph_id"],
            secondary_datasets=[secondary_dataset_1, secondary_dataset_2],
        )
    ]


@pytest.fixture
def secondary_dataset_configurations_response(ids):
    return {
        "projectId": ids["pid"],
        "config": [
            {
                "secondaryDatasets": [
                    {
                        "snapshotPolicy": "specified",
                        "identifier": ids["identifier_1"],
                        "catalogVersionId": ids["catalog_version_id_1"],
                        "catalogId": ids["catalog_id_1"],
                    },
                    {
                        "snapshotPolicy": "specified",
                        "identifier": ids["identifier_2"],
                        "catalogVersionId": ids["catalog_version_id_2"],
                        "catalogId": ids["catalog_id_2"],
                    },
                ],
                "featureEngineeringGraphId": ids["graph_id"],
            }
        ],
        "id": "5df109112ca582033ff44084",
    }


@responses.activate
def test_create_secondary_dataset_configurations(
    ids, dataset_configurations, secondary_dataset_configurations_response
):
    url = "https://host_name.com/projects/{}/secondaryDatasetsConfigurations/".format(ids["pid"])
    responses.add(
        responses.POST, url, status=201, body=json.dumps(secondary_dataset_configurations_response),
    )

    result = SecondaryDatasetConfigurations.create(ids["pid"], dataset_configurations)

    assert responses.calls[0].request.method == "POST"

    # verify request payload
    payload = request_body_to_json(responses.calls[0].request)
    conf = payload["conf"][0]
    assert conf["featureEngineeringGraphId"] == ids["graph_id"]
    dataset_confs = conf["secondaryDatasets"]
    expected_conf_sets = [d.to_dict() for d in dataset_configurations[0].secondary_datasets]
    assert len(dataset_confs) == len(expected_conf_sets)
    identifier = dataset_confs[0]["identifier"]
    expected_dataset = [d for d in expected_conf_sets if d["identifier"] == identifier][0]
    assert dataset_confs[0] == expected_dataset

    # verify response
    assert isinstance(result, SecondaryDatasetConfigurations)
    assert result.project_id == ids["pid"]
    dataset_conf = result.config[0]
    assert isinstance(dataset_conf, DatasetConfiguration)
    assert dataset_conf.feature_engineering_graph_id == ids["graph_id"]
    dataset = dataset_conf.secondary_datasets[0]
    assert isinstance(dataset, SecondaryDataset)
    identifier = dataset.identifier
    expected_dataset = [d for d in expected_conf_sets if d["identifier"] == identifier][0]
    assert dataset.catalog_id == expected_dataset["catalogId"]
    assert dataset.catalog_version_id == expected_dataset["catalogVersionId"]
    assert dataset.snapshot_policy == expected_dataset["snapshotPolicy"]


@responses.activate
def test_failure_in_create_secondary_dataset_configurations__status_error_code(
    ids, dataset_configurations
):
    url = "https://host_name.com/projects/{}/secondaryDatasetsConfigurations/".format(ids["pid"])
    responses.add(responses.POST, url, status=404, body="")

    with pytest.raises(errors.ClientError):
        SecondaryDatasetConfigurations.create(ids["pid"], dataset_configurations)


@responses.activate
def test_secondary_dataset_configurations_retrieve(
    ids, secondary_dataset_configurations_response,
):
    config_id = secondary_dataset_configurations_response["id"]
    pid = secondary_dataset_configurations_response["projectId"]
    dataset_configurations = secondary_dataset_configurations_response["config"]
    url = "https://host_name.com/projects/{}/secondaryDatasetsConfigurations/{}/".format(
        pid, config_id
    )
    responses.add(responses.GET, url, json=secondary_dataset_configurations_response)

    secondary_dataset_config = SecondaryDatasetConfigurations(id=config_id, project_id=pid).get()

    assert responses.calls[0].request.method == "GET"
    assert secondary_dataset_config.id == config_id
    assert secondary_dataset_config.project_id == pid
    assert len(secondary_dataset_config.config) == 1

    dataset_conf = secondary_dataset_config.config[0]
    assert isinstance(dataset_conf, DatasetConfiguration)
    assert dataset_conf.feature_engineering_graph_id == ids["graph_id"]
    dataset = dataset_conf.secondary_datasets[0]
    assert isinstance(dataset, SecondaryDataset)
    identifier = dataset.identifier
    expected_config = dataset_configurations[0]["secondaryDatasets"]
    expected_dataset = [d for d in expected_config if d["identifier"] == identifier][0]
    assert dataset.catalog_id == expected_dataset["catalogId"]
    assert dataset.catalog_version_id == expected_dataset["catalogVersionId"]
    assert dataset.snapshot_policy == expected_dataset["snapshotPolicy"]


@responses.activate
def test_secondary_dataset_configurations_deletion(secondary_dataset_configurations_response):
    config_id = secondary_dataset_configurations_response["id"]
    pid = secondary_dataset_configurations_response["projectId"]
    url = "https://host_name.com/projects/{}/secondaryDatasetsConfigurations/{}/".format(
        pid, config_id
    )
    responses.add(responses.GET, url, json=secondary_dataset_configurations_response)
    responses.add(responses.DELETE, url)
    secondary_dataset_config = SecondaryDatasetConfigurations(id=config_id, project_id=pid).get()
    secondary_dataset_config.delete()
    assert responses.calls[1].request.method == responses.DELETE
    assert responses.calls[1].request.url == url



import pytest

from datarobot.client import get_client
from tests.utils import SDKTestcase


class TestBuildRequestWithFile(SDKTestcase):
    def test_file_doesnt_exist(self):
        client = get_client()
        with self.assertRaises(ValueError):
            client.build_request_with_file(
                "PUT", "http://host-name/", "meh.csv", file_path="meh.csv"
            )

    def test_non_string_content(self):
        client = get_client()
        with self.assertRaises(AssertionError):
            client.build_request_with_file("PUT", "http://host-name/", "meh.csv", content=1234)


def test_strip_endpoint(client):
    url = "https://host_name.com/the/url/I/want/"
    stripped = client.strip_endpoint(url)
    assert stripped == "the/url/I/want/"


def test_strip_endpoint_with_training_slash(client):
    client.endpoint = "https://host_name.com/"
    url = "https://host_name.com/the/url/I/want/"
    stripped = client.strip_endpoint(url)
    assert stripped == "the/url/I/want/"


def test_strip_endpoint_invalid_url(client):
    url = "http://localhost/api/v2/wrong/prefix/on/this/url"
    with pytest.raises(ValueError):
        client.strip_endpoint(url)



import trafaret as t

from datarobot import CustomModelVersion, ExecutionEnvironment, ExecutionEnvironmentVersion
from datarobot._experimental.models.custom_training_model import CustomTrainingModel
from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2
from datarobot.utils.pagination import unpaginate


class CustomTrainingBlueprint(APIObject):
    """A custom training blueprint.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        blueprint id
    custom_model: dict
        dict with 2 keys: `id` and `name`,
        where `id` is the ID of the custom model
        and `name` is the model name
    custom_model_version: dict
        dict with 2 keys: `id` and `label`,
        where `id` is the ID of the custom model version
        and `label` is the version label
    execution_environment: dict
        dict with 2 keys: `id` and `name`,
        where `id` is the ID of the execution environment
        and `name` is the environment name
    execution_environment_version: dict
        dict with 2 keys: `id` and `label`,
        where `id` is the ID of the execution environment version
        and `label` is the version label
    """

    _path = "customTrainingBlueprints/"
    _converter = t.Dict(
        {
            t.Key("blueprint_id") >> "id": t.String(),
            t.Key("custom_model"): t.Dict({t.Key("id"): t.String(), t.Key("name"): t.String()}),
            t.Key("custom_model_version"): t.Dict(
                {t.Key("id"): t.String(), t.Key("label"): t.String()}
            ),
            t.Key("execution_environment"): t.Dict(
                {t.Key("id"): t.String(), t.Key("name"): t.String()}
            ),
            t.Key("execution_environment_version"): t.Dict(
                {t.Key("id"): t.String(), t.Key("label"): t.String()}
            ),
            t.Key("training_history"): t.List(t.Dict()),
        }
    ).ignore_extra("*")

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return encode_utf8_if_py2(u"{}({!r})".format(self.__class__.__name__, self.id))

    def _set_values(
        self,
        id,
        custom_model,
        custom_model_version,
        execution_environment,
        execution_environment_version,
        training_history,
        project_id=None,
        filenames=None,
    ):
        self.id = id
        self.custom_model = custom_model
        self.custom_model_version = custom_model_version
        self.execution_environment = execution_environment
        self.execution_environment_version = execution_environment_version
        self.training_history = training_history
        self.project_id = project_id
        self.filenames = filenames

    @classmethod
    def create(
        cls,
        custom_model_id,
        environment_id=None,
        custom_model_version_id=None,
        environment_version_id=None,
    ):
        """Create a custom learning blueprint.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model
        environment_id: Optional[str]
            the id of the execution environment
            If specified, the environment will be used "as-is"; if the custom model version
            has dependencies, they will not be installed at runtime.
        custom_model_version_id: Optional[str]
            the id of the custom model version
        environment_version_id: Optional[str]
            the id of the execution environment version. If version is none, what happens?


        if custom_model_id is provided, latest version will be used.
        if custom_model_version_id is provided, custom_model_id will be looked up
        One must be provided
        same for execution environments

        Returns
        -------
        CustomTrainingBlueprint
            created custom learning blueprint

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        if not custom_model_version_id:
            model = CustomTrainingModel.get(custom_model_id=custom_model_id)
            custom_model_version_id = model.latest_version.id

        payload = {
            "custom_model_id": custom_model_id,
            "custom_model_version_id": custom_model_version_id,
        }
        if environment_id:
            payload.update(
                {
                    "environment_id": environment_id,
                    "environment_version_id": environment_version_id,
                }
            )
        response = cls._client.post(cls._path, data=payload)
        return cls.from_server_data(response.json())

    @classmethod
    def create_from_scratch(cls, name, environment_dir, training_code_files, target_type):
        desc = "Generated from python client"
        # Make environment
        ee = ExecutionEnvironment.create(
            name=name + "-environment", description=desc, programming_language="python"
        )
        ExecutionEnvironmentVersion.create(
            str(ee.id), environment_dir, description=desc,
        )

        # Make custom model
        cm = CustomTrainingModel.create(
            name=name + "-model", target_type=target_type, description=desc
        )
        cmv = CustomModelVersion.create_clean(
            custom_model_id=cm.id, base_environment_id=ee.id, files=training_code_files
        )

        blueprint = cls.create(custom_model_id=cm.id, custom_model_version_id=cmv.id)
        blueprint.filenames = training_code_files
        return blueprint

    @classmethod
    def create_from_dropin(
        cls, model_name, dropin_env_id, target_type, training_code_files=None, folder_path=None
    ):
        new_custom_training_model = CustomTrainingModel.create(
            name=model_name, target_type=target_type
        )
        CustomModelVersion.create_clean(
            new_custom_training_model.id,
            base_environment_id=dropin_env_id,
            files=training_code_files,
            folder_path=folder_path,
        )
        blueprint = cls.create(custom_model_id=new_custom_training_model.id)
        blueprint.filenames = training_code_files
        return blueprint

    @classmethod
    def list(cls):
        """List custom learning blueprints.

        .. versionadded:: v2.21

        Returns
        -------
        List[CustomLearningBlueprint]
            a list of custom learning blueprints

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(cls._path, {}, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, blueprint_id):
        """Get custom learning blueprint by id.

        .. versionadded:: v2.21

        Parameters
        ----------
        blueprint_id: str
            the id of the custom learning blueprint

        Returns
        -------
        CustomTrainingBlueprint
            retrieved custom learning blueprint

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = "{}{}/".format(cls._path, blueprint_id)
        return cls.from_location(path)

    def with_updated_code(self, filenames=None):
        filenames_to_use = filenames
        if filenames is not None:
            self.filenames = filenames
        elif filenames is None and self.filenames is not None:
            filenames_to_use = self.filenames
        else:
            raise ValueError("Please provide all model filenames")

        cm_id = self.custom_model["id"]
        cmv = CustomModelVersion.create_clean(
            custom_model_id=cm_id,
            files=filenames_to_use,
            base_environment_id=self.execution_environment["id"],
        )
        self.custom_model_version = {"id": cmv.id, "label": cmv.label}
        payload = dict(custom_model_id=cm_id, custom_model_version_id=cmv.id)
        response = self._client.post(self._path, data=payload)
        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))
        return self



"""This module is not considered part of the public interface. As of 2.3, anything here
may change or be removed without warning."""

import os
import platform

import requests
from requests.adapters import HTTPAdapter
from requests_toolbelt.multipart.encoder import MultipartEncoder
import six
from six.moves.urllib_parse import urlparse
import trafaret as t
from urllib3 import Retry

from . import __version__, errors
from .enums import DEFAULT_TIMEOUT
from .utils import to_api


class RESTClientObject(requests.Session):
    """
    Parameters
        connect_timeout
            timeout for http request and connection
        headers
            headers for outgoing requests
    """

    @classmethod
    def from_config(cls, config):
        return cls(
            auth=config.token,
            endpoint=config.endpoint,
            connect_timeout=config.connect_timeout,
            verify=config.ssl_verify,
            user_agent_suffix=config.user_agent_suffix,
            max_retries=config.max_retries,
        )

    def __init__(
        self,
        auth=None,
        endpoint=None,
        connect_timeout=DEFAULT_TIMEOUT.CONNECT,
        verify=True,
        user_agent_suffix=None,
        max_retries=None,
    ):
        super(RESTClientObject, self).__init__()
        # Save the arguments needed to reconstruct a copy of this client
        # later in .copy()
        self._kwargs = {
            "auth": auth,
            "endpoint": endpoint,
            "connect_timeout": connect_timeout,
            "verify": verify,
            "user_agent_suffix": user_agent_suffix,
            "max_retries": max_retries,
        }
        # Note: As of 2.3, `endpoint` is required
        self.endpoint = endpoint
        self.domain = "{}://{}".format(
            urlparse(self.endpoint).scheme, urlparse(self.endpoint).netloc
        )
        self.token = auth
        if connect_timeout is None:
            connect_timeout = DEFAULT_TIMEOUT.CONNECT
        self.connect_timeout = connect_timeout
        self.user_agent_header = self._make_user_agent_header(suffix=user_agent_suffix)
        self.headers.update(self.user_agent_header)
        self.token_header = {"Authorization": "Token {}".format(self.token)}
        self.headers.update(self.token_header)
        self.verify = verify
        if max_retries is None:
            # by default, retry connect error but not read error for _all_ requests
            max_retries = Retry(connect=5, read=0, method_whitelist=False, backoff_factor=0.1)
        self.mount("http://", HTTPAdapter(max_retries=max_retries))
        self.mount("https://", HTTPAdapter(max_retries=max_retries))

    @staticmethod
    def _make_user_agent_header(suffix=None):
        py_version = platform.python_version()
        agent_components = [
            "DataRobotPythonClient/{}".format(__version__),
            "({} {} {})".format(platform.system(), platform.release(), platform.machine()),
            "Python-{}".format(py_version),
        ]
        if suffix:
            agent_components.append(suffix)
        return {"User-Agent": " ".join(agent_components)}

    def copy(self):
        """
        Get a copy of this RESTClientObject with the same configuration

        Returns
        -------
        client : RESTClientObject object.
        """
        return RESTClientObject(**self._kwargs)

    def _join_endpoint(self, url):
        return u"{}/{}".format(self.endpoint, url)

    def strip_endpoint(self, url):
        trailing = "" if self.endpoint.endswith("/") else "/"
        expected = "{}{}".format(self.endpoint, trailing)
        if not url.startswith(expected):
            raise ValueError(
                "unexpected url format: {} does not start with {}".format(url, expected)
            )
        return url.split(expected)[1]

    def request(self, method, url, join_endpoint=False, **kwargs):
        kwargs.setdefault("timeout", (self.connect_timeout, DEFAULT_TIMEOUT.READ))
        if not url.startswith("http") or join_endpoint:
            url = self._join_endpoint(url)
        response = super(RESTClientObject, self).request(method, url, **kwargs)
        if not response:
            handle_http_error(response, **kwargs)
        return response

    def get(self, url, params=None, **kwargs):
        return self.request("get", url, params=to_api(params), **kwargs)

    def post(self, url, data=None, keep_attrs=None, **kwargs):
        if data:
            kwargs["json"] = to_api(data, keep_attrs)
        return self.request("post", url, **kwargs)

    def patch(self, url, data=None, keep_attrs=None, **kwargs):
        if data:
            kwargs["json"] = to_api(data, keep_attrs=keep_attrs)
        return self.request("patch", url, **kwargs)

    def build_request_with_file(
        self,
        method,
        url,
        fname,
        form_data=None,
        content=None,
        file_path=None,
        filelike=None,
        read_timeout=DEFAULT_TIMEOUT.READ,
        file_field_name="file",
    ):
        """Build request with a file that will use special
        MultipartEncoder instance (lazy load file).


        This method supports uploading a file on local disk, string content,
        or a file-like descriptor. ``fname`` is a required parameter, and
        only one of the other three parameters can be provided.

        warning:: This function does not clean up its open files. This can
        lead to leaks. This isn't causing significant problems in the wild,
        and is non-trivial to solve, so we are leaving it as-is for now. This
        function uses a multipart encoder to gracefully handle large files
        when making the request. However, an encoder reference remains after
        the request has finished. So if the request body (i.e. the file) is
        accessed through the response after the response has been generated,
        the encoder uses the file descriptor opened here. If this descriptor
        is closed, then we raise an error when the encoder later tries to
        read the file again. This case exists in several of our tests, and
        may exist for users in the wild.

        Parameters
        ----------
        method : str.
            Method of request. This parameter is required, it can be
            'POST' or 'PUT' either 'PATCH'.
        url : str.
            Url that will be used it this request.
        fname : name of file
            This parameter is required, even when providing a file-like object
            or string content.
        content : str
            The content buffer of the file you would like to upload.
        file_path : str
            The path to a file on a local file system.
        filelike : file-like
            An open file descriptor to a file.
        read_timeout : float
            The number of seconds to wait after the server receives the file that we are
            willing to wait for the beginning of a response. Large file uploads may take
            significant time.
        file_field_name : str
            The name of the form field we will put the data into (Defaults to 'file').

        Returns
        -------
        response : response object.

        """
        bad_args_msg = (
            "Upload should be used either with content buffer "
            "or with path to file on local filesystem or with "
            "open file descriptor"
        )
        assert sum((bool(content), bool(file_path), bool(filelike))) == 1, bad_args_msg

        if file_path:
            if not os.path.exists(file_path):
                raise ValueError(u"Provided file does not exist {}".format(file_path))
            # See docstring for warning about unclosed file descriptors
            fields = {file_field_name: (fname, open(file_path, "rb"))}

        elif filelike:
            filelike.seek(0)
            fields = {file_field_name: (fname, filelike)}
        else:
            if not isinstance(content, six.binary_type):
                raise AssertionError("bytes type required in content")
            fields = {file_field_name: (fname, content)}

        form_data = form_data or {}
        data_for_encoder = to_api(form_data)
        data_for_encoder.update(fields)

        encoder = MultipartEncoder(fields=data_for_encoder)
        headers = {"Content-Type": encoder.content_type}
        return self.request(
            method, url, headers=headers, data=encoder, timeout=(self.connect_timeout, read_timeout)
        )


def _http_message(response):
    if response.status_code == 401:
        message = (
            "The server is saying you are not properly "
            "authenticated. Please make sure your API "
            "token is valid."
        )
    elif response.headers["content-type"] == "application/json":
        message = response.json()
    else:
        message = response.content.decode("ascii")[:79]
    return message


def handle_http_error(response, **kwargs):
    message = _http_message(response)
    if 400 <= response.status_code < 500:
        exception_type = errors.ClientError
        # One-off approach to raising special exception for now. We'll do something more
        # systematic when we have more of these:
        try:
            parsed_json = response.json()  # May raise error if response isn't JSON
            if parsed_json.get("errorName") == "JobAlreadyAdded":
                exception_type = errors.JobAlreadyRequested
        except (ValueError, AttributeError):
            parsed_json = {}
        template = "{} client error: {}"
        exc_message = template.format(response.status_code, message)
        raise exception_type(exc_message, response.status_code, json=parsed_json)
    else:
        template = "{} server error: {}"
        exc_message = template.format(response.status_code, message)
        raise errors.ServerError(exc_message, response.status_code)


class DataRobotClientConfig(object):
    """
    This class contains all of the client configuration variables that are known to
    the DataRobot client.

    Values are allowed to be None in this object. The __init__ of RESTClientObject will
    provide any defaults that should be applied if the user does not specify in the config
    """

    _converter = t.Dict(
        {
            t.Key("endpoint"): t.String(),
            t.Key("token"): t.String(),
            t.Key("connect_timeout", optional=True): t.Int(),
            t.Key("ssl_verify", optional=True): t.Or(t.Bool(), t.String()),
            t.Key("user_agent_suffix", optional=True): t.String(),
            t.Key("max_retries", optional=True): t.Int(),
        }
    ).allow_extra("*")
    _fields = {k.to_name or k.name for k in _converter.keys}

    def __init__(
        self,
        endpoint,
        token,
        connect_timeout=None,
        ssl_verify=True,
        user_agent_suffix=None,
        max_retries=None,
    ):
        self.endpoint = endpoint
        self.token = token
        self.connect_timeout = connect_timeout
        self.ssl_verify = ssl_verify
        self.user_agent_suffix = user_agent_suffix
        self.max_retries = max_retries

    @classmethod
    def from_data(cls, data):
        checked = {k: v for k, v in cls._converter.check(data).items() if k in cls._fields}
        return cls(**checked)



"""About datarobot
=========================
.. image:: https://img.shields.io/pypi/v/datarobot.svg
   :target: https://pypi.python.org/pypi/datarobot/
.. image:: https://img.shields.io/pypi/pyversions/datarobot.svg
.. image:: https://img.shields.io/pypi/status/datarobot.svg

DataRobot is a client library for working with the `DataRobot`_ platform API.

Installation
=========================
Python 2.7 and >= 3.4 are supported.
You must have a datarobot account.

::

   $ pip install datarobot

Usage
=========================
The library will look for a config file `~/.config/datarobot/drconfig.yaml` by default.
This is an example of what that config file should look like.

::

   token: your_token
   endpoint: https://app.datarobot.com/api/v2

Alternatively a global client can be set in the code.

::

   import datarobot as dr
   dr.Client(token='your_token', endpoint='https://app.datarobot.com/api/v2')

Alternatively environment variables can be used.

::

   export DATAROBOT_API_TOKEN='your_token'
   export DATAROBOT_ENDPOINT='https://app.datarobot.com/api/v2'

See `documentation`_ for example usage after configuring.

Tests
=========================
::

   $ py.test

.. _datarobot: http://datarobot.com
.. _documentation: https://datarobot-public-api-client.readthedocs-hosted.com
"""

import os
import re
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def run_tests(self):
        import shlex

        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


with open("datarobot/_version.py") as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError("Cannot find version information")

if os.environ.get("DR_CLIENT_BUILD_NUMBER"):
    version = "{version}+{build_number}".format(
        version=version, build_number=os.environ["DR_CLIENT_BUILD_NUMBER"],
    )

tests_require = [
    "mock==3.0.5",
    "pytest>=4.6,<5",  # 4.6 is last release supporting python2
    "pytest-cov",
    "responses>=0.9,<0.10",  # 0.10 changes the interface, will require test changes
]

dev_require = tests_require + [
    "flake8>=2.5.2,<3",
    "Sphinx==1.8.3",
    "sphinx_rtd_theme==0.1.9",
    "nbsphinx>=0.2.9,<1",
    "nbconvert==5.3.1",
    "numpydoc>=0.6.0",
    "tox",
    "jupyter_contrib_nbextensions",
    "tornado<6.0",
    'black==19.10b0; python_version >= "3.6"',
    'isort==5.4.2; python_version >= "3.6"',
]

example_require = [
    "jupyter<=5.0",
    "fredapi==0.4.0",
    "matplotlib>=2.1.0",
    "seaborn<=0.8",
    "sklearn<=0.18.2",
    "wordcloud<=1.3.1",
    "colour<=0.1.4",
]

release_require = ["zest.releaser[recommended]==6.22.0"]

setup(
    name="datarobot",
    version=version,
    description="This client library is designed to support the DataRobot API.",
    author="datarobot",
    author_email="support@datarobot.com",
    maintainer="datarobot",
    maintainer_email="info@datarobot.com",
    url="http://datarobot.com",
    license="Apache Software License",
    packages=[
        "datarobot",
        "datarobot._experimental",
        "datarobot._experimental.models",
        "datarobot.models",
        "datarobot.models.external_dataset_scores_insights",
        "datarobot.models.visualai",
        "datarobot.utils",
        "datarobot.helpers",
        "datarobot.ext",
    ],
    long_description=__doc__,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    install_requires=[
        "contextlib2>=0.5.5",
        "pandas>=0.15",
        "pyyaml>=3.11",
        "requests>=2.21",
        "requests_toolbelt>=0.6",
        "trafaret>=0.7,<2.0,!=1.1.0",
        "urllib3>=1.23",
        "attrs>=19.1.0,<20.0",
    ],
    tests_require=tests_require,
    extras_require={"dev": dev_require, "examples": example_require, "release": release_require},
    cmdclass={"test": PyTest},
)



import datetime

import pytest
import responses

from datarobot import FeatureEngineeringGraph, SharingAccess
from tests.utils import request_body_to_json


@pytest.fixture
def create_feature_engineering_graph_items():
    return [
        {
            "identifier": "user",
            "catalogVersionId": "5c88a37770fc42a2fcc62759",
            "snapshotPolicy": "latest",
        },
        {
            "identifier": "transaction",
            "catalogVersionId": "5c88a37770fc42a2fcc62760",
            "primaryTemporalKey": "date",
            "snapshotPolicy": "latest",
        },
    ]


@pytest.fixture
def create_feature_engineering_graph_relationships():
    return [
        {
            "table2Identifier": "user",
            "table1Keys": ["user_id", "dept_id"],
            "table2Keys": ["user_id", "dept_id"],
        },
        {
            "table1Identifier": "user",
            "table2Identifier": "transaction",
            "table1Keys": ["user_id"],
            "table2Keys": ["user_id"],
        },
    ]


@pytest.fixture
def feature_engineering_graph_response(
    create_feature_engineering_graph_items, create_feature_engineering_graph_relationships
):
    """ Payload to create a Feature Source via API
    """
    return {
        "id": "5a530498d5c1f302d6d176c8",
        "name": "Customer Features",
        "description": "Credit card customer features (using activity in last 2 months)",
        "tableDefinitions": create_feature_engineering_graph_items,
        "relationships": create_feature_engineering_graph_relationships,
        "featureDerivationWindowStart": -14,
        "featureDerivationWindowEnd": -1,
        "timeUnit": "DAY",
        "created": "2016-07-26T02:29:58.546312Z",
        "lastModified": "2016-07-26T02:29:58.546312Z",
        "creatorFullName": "abc@datarobot.com",
        "modifierFullName": "abc@datarobot.com",
        "creatorUserId": "5a530498d5c1f302d6d176c5",
        "lastModifiedUserId": "5a530498d5c1f302d6d176c5",
        "numberOfProjects": 2,
        "linkageKeys": ["user_id", "dept_id"],
    }


@pytest.fixture
def list_feature_engineering_graph_response(
    create_feature_engineering_graph_items, create_feature_engineering_graph_relationships
):
    return {
        "data": [
            {
                "id": "5a530498d5c1f302d6d176c8",
                "name": "Customer Features",
                "description": "Credit card customer features (using activity in last 2 months)",
                "tableDefinitions": create_feature_engineering_graph_items,
                "relationships": create_feature_engineering_graph_relationships,
                "featureDerivationWindowStart": -14,
                "featureDerivationWindowEnd": -1,
                "timeUnit": "DAY",
                "created": "2016-07-26T02:29:58.546312Z",
                "lastModified": "2016-07-26T02:29:58.546312Z",
                "creatorFullName": "abc@datarobot.com",
                "modifierFullName": "abc@datarobot.com",
                "creatorUserId": "5a530498d5c1f302d6d176c5",
                "lastModifiedUserId": "5a530498d5c1f302d6d176c5",
                "numberOfProjects": 2,
                "linkageKeys": ["user_id", "dept_id"],
            },
            {
                "id": "5a530498d5c1f302d6d175c8",
                "name": "Profile Features",
                "description": "Profile feature",
                "tableDefinitions": create_feature_engineering_graph_items,
                "relationships": create_feature_engineering_graph_relationships,
                "featureDerivationWindowStart": -10,
                "featureDerivationWindowEnd": -1,
                "timeUnit": "HOUR",
                "created": "2016-07-26T02:29:58.546312Z",
                "lastModified": "2016-07-26T02:29:58.546312Z",
                "creatorFullName": "abc@datarobot.com",
                "modifierFullName": "abc@datarobot.com",
                "creatorUserId": "5a530498d5c1f302d6d176c5",
                "lastModifiedUserId": "5a530498d5c1f302d6d176c5",
                "numberOfProjects": 2,
                "linkageKeys": ["user_id", "p_id"],
            },
        ]
    }


@responses.activate
def test_feature_engineering_graph_creation(
    feature_engineering_graph_response,
    create_feature_engineering_graph_items,
    create_feature_engineering_graph_relationships,
):
    responses.add(
        responses.POST,
        "https://host_name.com/featureEngineeringGraphs/",
        json=feature_engineering_graph_response,
    )

    feature_engineering_graph = FeatureEngineeringGraph()
    description = "Credit card customer features (using activity in last 2 months)"
    created_feg = feature_engineering_graph.create(
        name="Customer Features",
        description=description,
        table_definitions=create_feature_engineering_graph_items,
        relationships=create_feature_engineering_graph_relationships,
        time_unit="DAY",
        feature_derivation_window_start=-14,
        feature_derivation_window_end=-1,
    )
    assert responses.calls[0].request.method == "POST"
    assert created_feg.feature_engineering_graph_id == "5a530498d5c1f302d6d176c8"
    assert created_feg.name == "Customer Features"
    assert created_feg.description == description
    assert len(created_feg.table_definitions) == 2
    assert len(created_feg.relationships) == 2
    assert created_feg.time_unit == feature_engineering_graph_response["timeUnit"]
    assert created_feg.feature_derivation_window_start == -14
    assert created_feg.feature_derivation_window_end == -1
    assert isinstance(created_feg.created, datetime.datetime)
    assert isinstance(created_feg.last_modified, datetime.datetime)
    assert isinstance(created_feg.created, datetime.datetime)


@responses.activate
def test_feature_engineering_graph_replace(
    feature_engineering_graph_response,
    create_feature_engineering_graph_items,
    create_feature_engineering_graph_relationships,
):
    responses.add(
        responses.PUT,
        "https://host_name.com/featureEngineeringGraphs/F-ID/",
        status=200,
        json=feature_engineering_graph_response,
    )

    feature_engineering_graph = FeatureEngineeringGraph()
    description = "Credit card customer features (using activity in last 2 months)"
    updated_feg = feature_engineering_graph.replace(
        id="F-ID",
        name="Customer Features",
        description=description,
        table_definitions=create_feature_engineering_graph_items,
        relationships=create_feature_engineering_graph_relationships,
        time_unit="DAY",
        feature_derivation_window_start=-14,
        feature_derivation_window_end=-1,
        is_draft=False,
    )
    assert responses.calls[0].request.method == "PUT"
    payload = request_body_to_json(responses.calls[0].request)
    assert payload["name"] == "Customer Features"
    assert len(payload["tableDefinitions"]) == 2
    assert len(payload["relationships"]) == 2
    assert payload["timeUnit"] == "DAY"
    assert payload["featureDerivationWindowStart"] == -14
    assert payload["featureDerivationWindowEnd"] == -1

    assert updated_feg.feature_engineering_graph_id == "5a530498d5c1f302d6d176c8"
    assert updated_feg.name == "Customer Features"
    assert updated_feg.description == description
    assert len(updated_feg.table_definitions) == 2
    assert len(updated_feg.relationships) == 2
    assert updated_feg.time_unit == "DAY"
    assert updated_feg.feature_derivation_window_start == -14
    assert updated_feg.feature_derivation_window_end == -1
    assert isinstance(updated_feg.created, datetime.datetime)
    assert isinstance(updated_feg.last_modified, datetime.datetime)


@responses.activate
def test_feature_engineering_graph_updation():
    responses.add(
        responses.PATCH,
        "https://host_name.com/featureEngineeringGraphs/F-ID/",
        status=200,
        json={"name": "new-name", "description": "new-descrip"},
    )

    feature_engineering_graph = FeatureEngineeringGraph("F-ID")
    feature_engineering_graph.update(name="new-name", description="new-descrip")

    assert responses.calls[0].request.method == "PATCH"
    payload = request_body_to_json(responses.calls[0].request)
    assert payload["name"] == "new-name"
    assert payload["description"] == "new-descrip"


@responses.activate
def test_feature_engineering_graph_retrieve(feature_engineering_graph_response):
    graph_id = feature_engineering_graph_response["id"]
    responses.add(
        responses.GET,
        "https://host_name.com/featureEngineeringGraphs/{}/".format(graph_id),
        json=feature_engineering_graph_response,
    )

    feature_engineering_graph = FeatureEngineeringGraph.get(feature_engineering_graph_id=graph_id)

    assert responses.calls[0].request.method == "GET"
    assert feature_engineering_graph.feature_engineering_graph_id == graph_id
    assert feature_engineering_graph.name == "Customer Features"
    assert len(feature_engineering_graph.table_definitions) == 2
    assert len(feature_engineering_graph.relationships) == 2
    assert feature_engineering_graph.time_unit == "DAY"
    assert feature_engineering_graph.feature_derivation_window_start == -14
    assert feature_engineering_graph.feature_derivation_window_end == -1
    assert isinstance(feature_engineering_graph.created, datetime.datetime)
    assert isinstance(feature_engineering_graph.last_modified, datetime.datetime)
    assert isinstance(feature_engineering_graph.created, datetime.datetime)


@responses.activate
def test_feature_engineering_graph_list(list_feature_engineering_graph_response):
    responses.add(
        responses.GET,
        "https://host_name.com/featureEngineeringGraphs/",
        json=list_feature_engineering_graph_response,
    )

    graphs = FeatureEngineeringGraph.list()

    for graph, server_payload in zip(graphs, list_feature_engineering_graph_response["data"]):
        assert graph.feature_engineering_graph_id == server_payload["id"]
        assert graph.name == server_payload["name"]
        assert graph.description == server_payload["description"]
        assert len(graph.table_definitions) == 2
        assert len(graph.relationships) == 2
        fdw = server_payload["featureDerivationWindowStart"]
        assert graph.feature_derivation_window_start == fdw
        assert graph.feature_derivation_window_end == server_payload["featureDerivationWindowEnd"]
        assert graph.time_unit == server_payload["timeUnit"]
        assert graph.creator_full_name == server_payload["creatorFullName"]
        assert graph.modifier_full_name == server_payload["modifierFullName"]
        assert graph.number_of_projects == server_payload["numberOfProjects"]
        assert graph.linkage_keys == server_payload["linkageKeys"]


@responses.activate
def test_delete(feature_engineering_graph_response):
    graph_id = feature_engineering_graph_response["id"]
    responses.add(
        responses.GET,
        "https://host_name.com/featureEngineeringGraphs/{}/".format(graph_id),
        json=feature_engineering_graph_response,
    )
    delete_url = "https://host_name.com/featureEngineeringGraphs/{}/".format(graph_id)
    responses.add(responses.DELETE, delete_url)

    feature_engineering_graph = FeatureEngineeringGraph.get(graph_id)
    feature_engineering_graph.delete()

    assert responses.calls[1].request.method == responses.DELETE
    assert responses.calls[1].request.url == delete_url


@responses.activate
def test_share(feature_engineering_graph_response):
    graph = FeatureEngineeringGraph.from_server_data(feature_engineering_graph_response)
    request = SharingAccess("me@datarobot.com", "EDITOR")
    responses.add(
        responses.PATCH,
        "https://host_name.com/featureEngineeringGraphs/{}/accessControls/".format(
            graph.feature_engineering_graph_id
        ),
        status=200,
    )

    graph.share([request])

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {"permissions": [{"username": "me@datarobot.com", "role": "EDITOR"}]}


@responses.activate
def test_remove(feature_engineering_graph_response):
    graph = FeatureEngineeringGraph.from_server_data(feature_engineering_graph_response)
    request = SharingAccess("me@datarobot.com", None)
    responses.add(
        responses.PATCH,
        "https://host_name.com/featureEngineeringGraphs/{}/accessControls/".format(
            graph.feature_engineering_graph_id
        ),
        status=200,
    )

    graph.share([request])

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {"permissions": [{"username": "me@datarobot.com", "role": None}]}


@responses.activate
def test_get_access_list(feature_engineering_graph_response):
    graph = FeatureEngineeringGraph.from_server_data(feature_engineering_graph_response)
    access_list = {
        "data": [
            {
                "username": "me@datarobot.com",
                "userId": "1234deadbeeffeeddead4321",
                "role": "OWNER",
                "canShare": True,
            }
        ],
        "count": 1,
        "previous": None,
        "next": None,
    }

    access_record = access_list["data"][0]
    responses.add(
        responses.GET,
        "https://host_name.com/featureEngineeringGraphs/{}/accessControls/".format(
            graph.feature_engineering_graph_id
        ),
        json=access_list,
    )

    response = graph.get_access_list()
    assert len(response) == 1

    share_info = response[0]
    assert share_info.username == access_record["username"]
    assert share_info.user_id == access_record["userId"]
    assert share_info.role == access_record["role"]
    assert share_info.can_share == access_record["canShare"]



import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.sharing import SharingAccess
from datarobot.utils.pagination import unpaginate

from ..utils import encode_utf8_if_py2, parse_time


class FeatureEngineeringGraph(APIObject):
    """ A Feature Engineering Graph for the Project.
    A Feature Engineering Graph is graph which allow to specify relationships
    between two or more tables so it can automatically generate features from that

    Attributes
    ----------
    id : str
        the id of the created feature engineering graph
    name: str
        name of the feature engineering graph
    description: str
        description of the feature engineering graph
    created: datetime.datetime
        creation date of the feature engineering graph
    creator_user_id: str
        id of the user who created the feature engineering graph
    creator_full_name: str
        full name of the user who created the feature engineering graph
    last_modified: datetime.datetime
        last modification date of the feature engineering graph
    last_modified_user_id: str
        id of the user who last modified the feature engineering graph
    modifier_full_name: str
        full name of the user who last modified the feature engineering graph
    number_of_projects: int
        number of projects that are used in the feature engineering graph
    linkage_keys: list os str
        a list of strings specifying the name of the columns that link
        the feature engineering graph with the primary table.
    table_definitions: list
        each element is a table_definition for a table.
    relationships: list
        each element is a relationship between two tables
    time_unit: str, or None
        time unit of the feature derivation window. Supported
        values are MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR.
        If present, the feature engineering Graph will perform time-aware joins.
    feature_derivation_window_start: int, or None
        how many time_units of each table's primary temporal key into the past relative
        to the datetimePartitionColumn the feature derivation window should begin.
        Will be a negative integer,
        If present, the feature engineering Graph will perform time-aware joins.
    feature_derivation_window_end: int, or None
        how many timeUnits of each table's record
        primary temporal key into the past relative to the datetimePartitionColumn the feature
        derivation window should end.  Will be a non-positive integer, if present.
        If present, the feature engineering Graph will perform time-aware joins.
    is_draft: bool (default=True)
        a draft (is_draft=True) feature engineering graph can be updated,
        while a non-draft(is_draft=False) feature engineering graph is immutable

    The `table_defintions` structure is

    identifier: str
        alias of the table (used directly as part of the generated feature names)
    catalog_id: str, or None
        identifier of the catalog item
    catalog_version_id: str
        identifier of the catalog item version
    feature_list_id: str, or None
        identifier of the feature list. This decides which columns in the table are
        used for feature generation
    primary_temporal_key: str, or None
        name of the column indicating time of record creation
    snapshot_policy: str
        policy to use  when creating a project or making predictions.
        Must be one of the following values:
        'specified': Use specific snapshot specified by catalogVersionId
        'latest': Use latest snapshot from the same catalog item
        'dynamic': Get data from the source (only applicable for JDBC datasets)
    feature_lists: list
        list of feature list info
    data_source: dict
        data source info if the table is from data source
    is_deleted: bool or None
        whether the table is deleted or not

    The `relationship` structure is

    table1_identifier: str or None
        identifier of the first table in this relationship.
        This is specified in the indentifier field of table_definition structure.
        If None, then the relationship is with the primary dataset.
    table2_identifier: str
        identifier of the second table in this relationship.
        This is specified in the identifier field of table_definition schema.
    table1_keys: list of str (max length: 10 min length: 1)
        column(s) from the first table which are used to join to the second table
    table2_keys: list of str (max length: 10 min length: 1)
        column(s) from the second table that are used to join to the first table

    The `feature list info` structure is

    id : str
        the id of the featurelist
    name : str
        the name of the featurelist
    features : list of str
        the names of all the Features in the featurelist
    dataset_id : str
        the project the featurelist belongs to
    creation_date : datetime.datetime
        when the featurelist was created
    user_created : bool
        whether the featurelist was created by a user or by DataRobot automation
    created_by: str
        the name of user who created it
    description : str
        the description of the featurelist.  Can be updated by the user
        and may be supplied by default for DataRobot-created featurelists.
    dataset_id: str
        dataset which is associated with the feature list
    dataset_version_id: str or None
        version of the dataset which is associated with feature list.
        Only relevant for Informative features

    The `data source info` structured is

    data_store_id: str
        the id of the data store.
    data_store_name : str
        the user-friendly name of the data store.
    url : str
        the url used to connect to the data store.
    dbtable : str
        the name of table from the data store.
    schema: str
        schema definition of the table from the data store
    """

    _path = "featureEngineeringGraphs/"
    data_source_trafaret = t.Dict(
        {
            t.Key("data_store_name"): t.String,
            t.Key("data_store_id"): t.String,
            t.Key("url"): t.String,
            t.Key("dbtable"): t.String | t.Null,
            t.Key("schema", optional=True): t.String | t.Null,
        }
    )

    feature_list_info = t.Dict(
        {
            t.Key("id"): t.String,
            t.Key("features"): t.List(t.String),
            t.Key("name"): t.String,
            t.Key("description"): t.String,
            t.Key("dataset_id"): t.String,
            t.Key("user_created"): t.Bool,
            t.Key("creation_date"): parse_time,
            t.Key("created_by"): t.String,
            t.Key("dataset_version_id", optional=True): t.String,
        }
    )

    table_definition_trafaret = t.Dict(
        {
            t.Key("identifier"): t.String(min_length=3, max_length=20),
            t.Key("catalog_version_id"): t.String,
            t.Key("catalog_id", optional=True): t.String,
            t.Key("primary_temporal_key", optional=True): t.String | t.Null,
            t.Key("feature_list_id", optional=True): t.String | t.Null,
            t.Key("feature_lists", optional=True): t.List(feature_list_info),
            t.Key("snapshot_policy", optional=True, default="latest"): t.Enum(
                "latest", "specified", "dynamic"
            ),
            t.Key("data_source", optional=True): data_source_trafaret | t.Null,
            t.Key("is_deleted", optional=True): t.Bool | t.Null,
        }
    )

    relationships_trafaret = t.Dict(
        {
            t.Key("table1_identifier", optional=True): t.String | t.Null,
            t.Key("table2_identifier"): t.String,
            t.Key("table1_keys"): t.List(t.String, min_length=1, max_length=10),
            t.Key("table2_keys"): t.List(t.String, min_length=1, max_length=10),
        }
    )
    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("name"): t.String(max_length=100, allow_blank=False),
            t.Key("description"): t.String(max_length=300, allow_blank=False),
            t.Key("created"): parse_time,
            t.Key("last_modified"): parse_time,
            t.Key("creator_full_name"): t.String,
            t.Key("modifier_full_name"): t.String,
            t.Key("creator_user_id"): t.String,
            t.Key("last_modified_user_id"): t.String,
            t.Key("number_of_projects"): t.Int,
            t.Key("linkage_keys"): t.List(t.String),
            t.Key("table_definitions"): t.List(table_definition_trafaret, min_length=1),
            t.Key("relationships", optional=True): t.List(relationships_trafaret),
            t.Key("time_unit", optional=True): t.String,
            t.Key("feature_derivation_window_start", optional=True): t.Int(lt=0),
            t.Key("feature_derivation_window_end", optional=True): t.Int(lte=0),
            t.Key("is_draft", optional=True): t.Bool(),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id=None,
        name=None,
        description=None,
        created=None,
        last_modified=None,
        creator_full_name=None,
        modifier_full_name=None,
        creator_user_id=None,
        last_modified_user_id=None,
        number_of_projects=None,
        linkage_keys=None,
        table_definitions=None,
        relationships=None,
        time_unit=None,
        feature_derivation_window_start=None,
        feature_derivation_window_end=None,
        is_draft=True,
    ):
        self.feature_engineering_graph_id = id
        self.name = name
        self.description = description
        self.created = created
        self.last_modified = last_modified
        self.creator_full_name = creator_full_name
        self.modifier_full_name = modifier_full_name
        self.creator_user_id = creator_user_id
        self.last_modified_user_id = last_modified_user_id
        self.number_of_projects = number_of_projects
        self.linkage_keys = linkage_keys
        self.table_definitions = table_definitions
        self.relationships = relationships
        self.time_unit = time_unit
        self.feature_derivation_window_start = feature_derivation_window_start
        self.feature_derivation_window_end = feature_derivation_window_end
        self.is_draft = is_draft

    def __repr__(self):
        return encode_utf8_if_py2(u"{}()".format(self.__class__.__name__))

    def _build_url(self, feature_engineering_graph_id):
        return "{}{}/".format(self._path, feature_engineering_graph_id)

    @classmethod
    def _create_payload(
        cls,
        name,
        description,
        table_definitions,
        relationships,
        time_unit,
        feature_derivation_window_start,
        feature_derivation_window_end,
        is_draft,
    ):
        """
        Create the payload to create feature engineering graph

        Parameters
        ----------
        name: str
            name of the feature engineering graph
        description: str
            description of the fearture engineering graph
        table_definitions: list
            each element is a table_definition for a table.
        relationships: list
            each element is a relationship between two tables
        time_unit: str or None
            time unit of the feature derivation window. Supported
            values are MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR.
            If present, the feature engineering Graph will perform time-aware joins.
        feature_derivation_window_start: int, or None
            how many time_units of each table's primary temporal key into the past relative
            to the datetimePartitionColumn the feature derivation window should begin.
            Will be a negative integer,
            If present, the feature engineering Graph will perform time-aware joins.
        feature_derivation_window_end: int, or None
            how many timeUnits of each table's record
            primary temporal key into the past relative to the datetimePartitionColumn the feature
            derivation window should end.  Will be a non-positive integer, if present.
            If present, the feature engineering Graph will perform time-aware joins.
        is_draft: bool (default=True)
            a draft (is_draft=True) feature engineering graph can be updated,
            while a non-draft(is_draft=False) feature engineering graph is immutable
        Returns
        -------
        payload: dict
            payload for the feature engineering graph
        """
        payload = {
            "name": name,
            "description": description,
            "tableDefinitions": table_definitions,
            "relationships": relationships,
        }

        if time_unit is not None:
            payload["timeUnit"] = time_unit

        if feature_derivation_window_start is not None:
            payload["featureDerivationWindowStart"] = feature_derivation_window_start

        if feature_derivation_window_end is not None:
            payload["featureDerivationWindowEnd"] = feature_derivation_window_end

        payload["isDraft"] = is_draft
        return payload

    @classmethod
    def create(
        cls,
        name,
        description,
        table_definitions,
        relationships,
        time_unit=None,
        feature_derivation_window_start=None,
        feature_derivation_window_end=None,
        is_draft=True,
    ):
        """
        Create a feature engineering graph.

        Parameters
        ----------
        name : str
            the name of the feature engineering graph
        description : str
            the description of the feature engineering graph
        table_definitions: list of dict
            each element is a TableDefinition for a table.
            The `TableDefinition` schema is

            identifier: str
                alias of the table (used directly as part of the generated feature names)
            catalog_id: str, or None
                identifier of the catalog item
            catalog_version_id: str
                identifier of the catalog item version
            feature_list_id: str, or None
                identifier of the feature list. This decides which columns in the table are
                used for feature generation
            primary_temporal_key: str, or None
                name of the column indicating time of record creation
            snapshot_policy: str
                policy to use  when creating a project or making predictions.
                Must be one of the following values:
                'specified': Use specific snapshot specified by catalogVersionId
                'latest': Use latest snapshot from the same catalog item
                'dynamic': Get data from the source (only applicable for JDBC datasets)
        relationships: list of dict
            each element is a Relationship between two tables
            The `Relationship` schema is

            table1_identifier: str or None
                identifier of the first table in this relationship.
                This is specified in the indentifier field of table_definition structure.
                If None, then the relationship is with the primary dataset.
            table2_identifier: str
                identifier of the second table in this relationship.
                This is specified in the identifier field of table_definition schema.
            table1_keys: list of str (max length: 10 min length: 1)
                column(s) from the first table which are used to join to the second table
            table2_keys: list of str (max length: 10 min length: 1)
                column(s) from the second table that are used to join to the first table
        time_unit: str, or None
            time unit of the feature derivation window. Supported
            values are MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR.
            If present, the feature engineering Graph will perform time-aware joins.
        feature_derivation_window_start: int, or None
            how many time_units of each table's primary temporal key into the past relative
            to the datetimePartitionColumn the feature derivation window should begin.
            Will be a negative integer,
            If present, the feature engineering Graph will perform time-aware joins.
        feature_derivation_window_end: int, or None
            how many timeUnits of each table's record
            primary temporal key into the past relative to the datetimePartitionColumn the feature
            derivation window should end.  Will be a non-positive integer, if present.
            If present, the feature engineering Graph will perform time-aware joins.
        is_draft: bool (default=True)
            a draft (is_draft=True) feature engineering graph can be updated,
            while a non-draft(is_draft=False) feature engineering graph is immutable

        Returns
        -------
        feature_engineering_graphs: FeatureEngineeringGraph
            the created feature engineering graph
        """
        payload_data = cls._create_payload(
            name=name,
            description=description,
            table_definitions=table_definitions,
            relationships=relationships,
            time_unit=time_unit,
            feature_derivation_window_start=feature_derivation_window_start,
            feature_derivation_window_end=feature_derivation_window_end,
            is_draft=is_draft,
        )

        response = cls._client.post(cls._path, data=payload_data).json()
        return FeatureEngineeringGraph.from_server_data(response)

    def replace(
        self,
        id,
        name,
        description,
        table_definitions,
        relationships,
        time_unit=None,
        feature_derivation_window_start=None,
        feature_derivation_window_end=None,
        is_draft=True,
    ):
        """
        Replace a feature engineering graph.

        Parameters
        ----------
        id : str
            the id of the created feature engineering graph
        name : str
            the name of the feature engineering graph
        description : str
            the description of the feature engineering graph
        items: list of dict
            each element is a TableDefinition for a table.
            The `TableDefinition` schema is

            identifier: str
                alias of the table (used directly as part of the generated feature names)
            catalog_id: str, or None
                identifier of the catalog item
            catalog_version_id: str
                identifier of the catalog item version
            feature_list_id: str, or None
                identifier of the feature list. This decides which columns in the table are
                used for feature generation
            primary_temporal_key: str, or None
                name of the column indicating time of record creation
            snapshot_policy: str
                policy to use  when creating a project or making predictions.
                Must be one of the following values:
                'specified': Use specific snapshot specified by catalogVersionId
                'latest': Use latest snapshot from the same catalog item
                'dynamic': Get data from the source (only applicable for JDBC datasets)
        relationships: list of dict
            each element is a Relationship between two tables
            The `Relationship` schema is

            table1_identifier: str or None
                identifier of the first table in this relationship.
                This is specified in the indentifier field of table_definition structure.
                If None, then the relationship is with the primary dataset.
            table2_identifier: str
                identifier of the second table in this relationship.
                This is specified in the identifier field of table_definition schema.
            table1_keys: list of str (max length: 10 min length: 1)
                column(s) from the first table which are used to join to the second table
            table2_keys: list of str (max length: 10 min length: 1)
                column(s) from the second table that are used to join to the first table
        time_unit: str, or None
            time unit of the feature derivation window. Supported
            values are MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR.
            If present, the feature engineering Graph will perform time-aware joins.
        feature_derivation_window_start: int, or None
            how many time_units of each table's primary temporal key into the past relative
            to the datetimePartitionColumn the feature derivation window should begin.
            Will be a negative integer,
            If present, the feature engineering Graph will perform time-aware joins.
        feature_derivation_window_end: int, or None
            how many timeUnits of each table's record
            primary temporal key into the past relative to the datetimePartitionColumn the feature
            derivation window should end.  Will be a non-positive integer, if present.
            If present, the feature engineering Graph will perform time-aware joins.
        is_draft: bool (default=True)
            a draft (is_draft=True) feature engineering graph can be updated,
            while a non-draft(is_draft=False) feature engineering graph is immutable

        Returns
        -------
        feature_engineering_graphs: FeatureEngineeringGraph
            the updated feature engineering graph
        """
        payload_data = self._create_payload(
            name=name,
            description=description,
            table_definitions=table_definitions,
            relationships=relationships,
            time_unit=time_unit,
            feature_derivation_window_start=feature_derivation_window_start,
            feature_derivation_window_end=feature_derivation_window_end,
            is_draft=is_draft,
        )

        response = self._client.put(
            self._build_url(feature_engineering_graph_id=id), json=payload_data
        ).json()
        return FeatureEngineeringGraph.from_server_data(response)

    def update(self, name, description):
        """
        Update the Feature engineering graph name and description.

        Parameters
        ----------
        name : str
            the name of the feature engineering graph
        description : str
            the description of the feature engineering graph
        """
        payload = {"name": name, "description": description}
        build_url = self._build_url(feature_engineering_graph_id=self.feature_engineering_graph_id)
        self._client.patch(build_url, data=payload)

    @classmethod
    def get(cls, feature_engineering_graph_id):
        """
        Retrieve a single feature engineering graph

        Parameters
        ----------
        feature_engineering_graph_id : str
            The ID of the feature engineering graph to retrieve.

        Returns
        -------
        feature_engineering_graph : FeatureEngineeringGraph
            The requested feature engineering graph
        """
        return cls.from_location("{}{}/".format(cls._path, feature_engineering_graph_id))

    @classmethod
    def list(cls, project_id=None, secondary_dataset_id=None, include_drafts=None):
        """
        Returns list of feature engineering graphs.

        Parameters
        ----------
        project_id: str, optional
            The Id of project to filter the feature engineering graph list for returning
            only those feature engineering Graphs which are related to this project
            If not specified, it will return all the feature engineering graphs
            irrespective of the project
        secondary_dataset_id: str, optional
            ID of the dataset to filter feature engineering graphs which use the
            dataset as the secondary dataset
            If not specified, return all the feature engineering graphs without filtering on
            secondary dataset id.
        include_drafts: bool (default=False)
            include draft feature engineering graphs
            If `True`, return all the draft (mutable) as well as non-draft (immutable)
            feature engineering graphs

        Returns
        -------
        feature_engineering_graphs : list of FeatureEngineeringGraph instances
            a list of available feature engineering graphs.
        """
        kwargs = {}
        if project_id:
            kwargs["projectId"] = project_id
        if secondary_dataset_id:
            kwargs["secondaryDatasetId"] = secondary_dataset_id
        if include_drafts:
            kwargs["includeDrafts"] = include_drafts

        r_data = cls._client.get(cls._path, params=kwargs).json()
        return [FeatureEngineeringGraph.from_server_data(item) for item in r_data["data"]]

    def delete(self):
        """
        Delete the Feature Engineering Graph
        """
        self._client.delete("{}{}/".format(self._path, self.feature_engineering_graph_id))

    def share(self, access_list):
        """
        Modify the ability of users to access this feature engineering graph

        Parameters
        ----------
        access_list : list of :class:`SharingAccess <datarobot.SharingAccess>`
            the modifications to make.

        Raises
        ------
        datarobot.ClientError :
            if you do not have permission to share this feature engineering graph or
            if the user you're sharing with doesn't exist

        """
        payload = {"permissions": [access.collect_payload() for access in access_list]}
        self._client.patch(
            "{}{}/accessControls/".format(self._path, self.feature_engineering_graph_id),
            data=payload,
            keep_attrs={"role"},
        )

    def get_access_list(self):
        """
        Retrieve what users have access to this feature engineering graph

        Returns
        -------
        list of :class:`SharingAccess <datarobot.SharingAccess>`
        """
        url = "{}{}/accessControls/".format(self._path, self.feature_engineering_graph_id)
        return [
            SharingAccess.from_server_data(datum) for datum in unpaginate(url, {}, self._client)
        ]
